"""API methods for working with single resources."""
import json
from typing import Optional
from urllib.parse import quote_plus

import requests
from cmem.cmempy import config

from cmem.cmempy.api import request, send_request


def get_resource_uri(
    project_name: Optional[str] = None,
    resource_name: Optional[str] = None,
    sub_path: Optional[str] = None,
) -> str:
    """
    Get the endpoint URI pattern for a resource OR get
    the full URI in case project and resource is given.

    Args:
        project_name: the project ID as a string
        resource_name: the resource path
        sub_path: None, "metadata" or "usage"

    Note: resource_name will be URL encoded before request.
    """
    path = "/workspace/projects/{}/files?path={}"
    # sub_path can be used to request a sub-API of the files API
    if sub_path:
        path = path.replace("files?", f"files/{sub_path}?")
    path = config.get_di_api_endpoint() + path
    if not project_name and not resource_name:
        return str(path)
    if project_name and resource_name:
        return str(path.format(project_name, quote_plus(resource_name)))
    raise ValueError(
        "Either give project_name AND resource_name or give none of them"
        " and get the pattern only."
    )


def get_resource(project_name, resource_name):
    """GET resource.

    Args:
        project_name (str): The project ID in the workspace.
        resource_name (str): The resource ID/name in the workspace.

    Returns:
        requests.Response object
    """
    resource_url = get_resource_uri(
        project_name=project_name, resource_name=resource_name
    )
    response = send_request(resource_url, method="GET")
    return response


def resource_exist(project_name, resource_name):
    """Check if a resource exist.

    A return value of true means the resource exists. A return value of true
    means the resource does not exists.

    Args:
        project_name (str): The project ID in the workspace.
        resource_name (str): The resource ID/name in the workspace.

    Returns:
        bool
    """
    try:
        get_resource_metadata(project_name, resource_name)
        return True
    except requests.exceptions.HTTPError as error:
        if error.response.status_code == 404:
            return False
        raise error


def get_resource_response(project_name, resource_name):
    """GET resource as streamable request response.

    Args:
        project_name (str): The project ID in the workspace.
        resource_name (str): The resource ID/name in the workspace.

    Returns:
        requests.Response object
    """
    resource_url = get_resource_uri(
        project_name=project_name, resource_name=resource_name
    )
    return request(resource_url, method="GET", stream=True)


def make_new_resource(project_name, resource_name, data=None, files=None):
    """PUT create new resource.

    Args:
        project_name (str): The project ID in the workspace.
        resource_name (str): The resource ID/name in the workspace.
        data (dict): Dictionary (see requests.request) for the body
        files (dict): Dictionary (see requests.request) for multipart upload
    """
    resource_url = get_resource_uri(
        project_name=project_name, resource_name=resource_name
    )
    send_request(
        resource_url,
        method="PUT",
        data=data,
        files=files,
    )


def create_resource(project_name, resource_name, file_resource=None, replace=False):
    """Create a new resource (streamed).

    Args:
        project_name (str): The project ID in the workspace.
        resource_name (str): The resource ID/name in the workspace.
        file_resource (file stream): Already opened byte file stream
        replace (bool): Replace resource if needed.

    Returns:
        requests.Response object

    Raises:
        ValueError: missing parameter
        ValueError: Resource exist and no replace enabled
    """
    if not file_resource:
        raise ValueError("Parameter file_name is needed.")
    if not replace and resource_exist(project_name, resource_name):
        raise ValueError(
            f"Resource {resource_name} already exists " f"in project {project_name}."
        )
    resource_url = get_resource_uri(
        project_name=project_name, resource_name=resource_name
    )
    # https://requests.readthedocs.io/en/latest/user/advanced/#streaming-uploads
    with file_resource as file:
        response = request(
            resource_url,
            method="PUT",
            stream=True,
            data=file,
        )
    return response


def delete_resource(project_name, resource_name):
    """DELETE remove existing resource.

    Args:
        project_name (str): The project ID in the workspace.
        resource_name (str): The resource ID/name in the workspace.
    """
    resource_url = get_resource_uri(
        project_name=project_name, resource_name=resource_name
    )
    send_request(resource_url, method="DELETE")


def get_resource_metadata(project_name, resource_name):
    """GET retrieve resource metadata.

    Args:
        project_name (str): The project ID in the workspace.
        resource_name (str): The resource ID/name in the workspace.

    Returns:
        Depends on what json.loads gives back
    """
    resource_url = get_resource_uri(
        project_name=project_name, resource_name=resource_name, sub_path="metadata"
    )
    response = send_request(resource_url, method="GET")
    return json.loads(response.decode("utf-8"))


def get_resource_usage_data(project_name, resource_name):
    """GET retrieve the usage data of a resource.

    Args:
        project_name (str): The project ID in the workspace.
        resource_name (str): The resource ID/name in the workspace.

    Returns:
        Depends on what json.loads gives back
    """
    resource_url = get_resource_uri(
        project_name=project_name, resource_name=resource_name, sub_path="usage"
    )
    response = send_request(resource_url, method="GET")
    return json.loads(response.decode("utf-8"))



"""API for working with several resources."""
import json

from cmem.cmempy import config

from cmem.cmempy.api import send_request
from cmem.cmempy.workspace.projects.project import get_projects


def get_resources_uri():
    """Get endpoint URI pattern for resource list."""
    path = "/workspace/projects/{}/resources"
    return config.get_di_api_endpoint() + path


def get_resources(project_name):
    """GET retrieve list of resources."""
    response = send_request(get_resources_uri().format(project_name), method="GET")
    return json.loads(response.decode("utf-8"))


def get_all_resources():
    """Get all resources of all projects."""
    resources = []
    for project_id in [_["name"] for _ in get_projects()]:
        for resource in list(get_resources(project_id)):
            resource["project"] = project_id
            resource["id"] = project_id + ":" + resource["fullPath"]
            resources.append(resource)
    return resources



"""API for searching items in the data integration workspace."""

import json

from cmem.cmempy import config

from cmem.cmempy.api import send_request


def get_search_items_uri():
    """Get endpoint URL for the search items API."""
    path = "/api/workspace/searchItems"
    return config.get_di_api_endpoint() + path


# pylint: disable-msg=too-many-arguments
def list_items(
    limit=1000000,
    offset=0,
    project=None,
    text_query="",
    item_type="",
    facets=None,
    add_task_parameter=True,
):
    """GET retrieve custom task for the project."""
    data = json.dumps(
        {
            "limit": limit,
            "offset": offset,
            "itemType": item_type,
            "project": project,
            "textQuery": text_query,
            "facets": facets,
            "addTaskParameters": add_task_parameter,
        }
    )
    response = send_request(
        get_search_items_uri(),
        method="POST",
        headers={"Content-Type": "application/json"},
        data=data,
    )
    return json.loads(response.decode("utf-8"))



"""API for working with several workflows."""
import json

from cmem.cmempy import config

from cmem.cmempy.api import get_json, send_request


def get_workflow_uri():
    """Get endpoint URI pattern for a workflow."""
    path = "/workflow/workflows/{}"
    return config.get_di_api_endpoint() + path


def get_workflows(project_name):
    """GET retrieve list of workflows."""
    response = send_request(get_workflow_uri().format(project_name), method="GET")
    return json.loads(response.decode("utf-8"))


def get_resource_based_dataset_types():
    """Get a list of dataset types for all resources based."""

    return get_json(
        config.get_di_api_endpoint() + "/api/core/datasets/resourceBased", method="GET"
    )



"""API for health and version information."""
import json
import re

from requests import HTTPError

from cmem.cmempy import config
from cmem.cmempy.api import get_json, send_request
from cmem.cmempy.queries import SparqlQuery

SHAPE_CATALOG_VERSION_QUERY = """
PREFIX owl: <http://www.w3.org/2002/07/owl#>
PREFIX : <https://vocab.eccenca.com/shacl/>
SELECT ?version
FROM :
WHERE {
  : owl:versionInfo ?version
}
ORDER BY ASC(?version)
"""


def get_explore_health_endpoint():
    """Get explore health endpoint."""
    return get_dp_health_endpoint()


def get_dp_health_endpoint():
    """Get DataPlatform health endpoint."""
    return config.get_dp_api_endpoint() + "/actuator/health"


def get_dp_info_endpoint():
    """Get DataPlatform version endpoint."""
    return config.get_dp_api_endpoint() + "/actuator/info"


def get_di_health_endpoint():
    """Get DataPlatform health endpoint."""
    return config.get_di_api_endpoint() + "/health"


def get_di_version_endpoint():
    """Get DataPlatform version endpoint."""
    return config.get_di_api_endpoint() + "/version"


def get_dm_version_endpoint():
    """Get DataManager version endpoint."""
    return config.get_cmem_base_uri() + "/version.html"


def get_shape_catalog_version():
    """GET version of the ShapeCatalog."""
    try:
        results = SparqlQuery(SHAPE_CATALOG_VERSION_QUERY).get_json_results()[
            "results"
        ]["bindings"]
        if len(results) > 0:
            return str(results[0]["version"]["value"])
    except (KeyError, HTTPError):
        pass
    return "UNKNOWN"


def get_dm_version():
    """GET version of DataManager."""
    url = get_dm_version_endpoint()
    try:
        response = send_request(url).decode()
    except HTTPError:
        return "ERROR"
    try:
        response = re.findall(r"v[0-9]+\..*", response)[0]
        return response
    except Exception:  # pylint: disable=broad-except
        return "UNKNOWN"


def get_explore_version():
    """GET version of Explore."""
    return get_dp_version()


def get_dp_version():
    """GET version of DataPlatform."""
    response = None
    url = get_dp_info_endpoint()
    try:
        response = send_request(url)
    except Exception:  # pylint: disable=broad-except
        response = None
        # TODO: checking health status needs to be improved
    if response is None:
        url = url.replace("/actuator", "")
        response = send_request(url)
    result = json.loads(response)
    return result["version"]


def dp_is_healthy():
    """Check status of DataIntegration."""
    url = get_dp_health_endpoint()
    try:
        response = send_request(url)
    except Exception:  # pylint: disable=broad-except
        response = None
        # TODO: checking health status needs to be improved
    if response is None:
        url = url.replace("/actuator", "")
        response = send_request(url)
    result = json.loads(response)
    if result["status"] == "UP":
        return True
    return False


def get_di_version():
    """GET version of DataIntegration."""
    response = send_request(get_di_version_endpoint())
    return response.decode("utf-8")


def di_is_healthy():
    """Check status of DataIntegration."""
    try:
        result = json.loads(send_request(get_di_health_endpoint()))
    except ValueError:
        return False
    if result["status"] == "UP":
        return True
    return False


def get_complete_status_info() -> dict:
    """Collect all status info and provide proper dictionary."""
    info = {
        "build": {"version": "UNKNOWN", "healthy": "UNKNOWN"},
        "explore": {"version": "UNKNOWN", "healthy": "UNKNOWN"},
        "shapes": {"version": "UNKNOWN", "healthy": "UNKNOWN"},
        "store": {"type": "STORE", "version": "UNKNOWN", "healthy": "UNKNOWN"},
    }
    e_prefix = "Error: "
    try:
        info["build"]["version"] = get_di_version()
        info["build"]["healthy"] = "UP" if di_is_healthy() else "DOWN"
    except Exception as error:  # pylint: disable=broad-except
        info["build"]["error"] = f"{e_prefix}{error}"
        info["build"]["healthy"] = "DOWN"
    try:
        info["explore"]["info"] = get_explore_info()  # type: ignore
        info["explore"]["health"] = get_json(get_explore_health_endpoint())
        info["explore"]["version"] = info["explore"]["info"]["version"]  # type: ignore
        info["explore"]["healthy"] = info["explore"]["health"]["status"]  # type: ignore
        if "store" in info["explore"]["info"]:
            info["store"]["type"] =\
                info["explore"]["info"]["store"]["type"]  # type: ignore
            # fmt: off
            info["store"]["version"] = (
                info["explore"]["info"]["store"]["version"]  # type: ignore
            )
            # fmt: on
        info["store"]["healthy"] = info["explore"]["healthy"]
    except Exception as error:  # pylint: disable=broad-except
        info["explore"]["error"] = f"{e_prefix}{error}"
        info["explore"]["healthy"] = "DOWN"
    try:
        info["shapes"]["version"] = get_shape_catalog_version()
        if info["shapes"]["version"] == "UNKNOWN":
            info["shapes"]["healthy"] = "DOWN"
        else:
            info["shapes"]["healthy"] = "UP"
    except Exception as error:  # pylint: disable=broad-except
        info["shapes"]["error"] = f"{e_prefix}{error}"
    # combine overall.status
    info["overall"] = {
        "healthy": "UP" if all(
            component["healthy"] == "UP" for component in info.values()
        ) else "DOWN"
    }
    return info


def get_explore_info() -> dict:
    """Get explore info."""
    return get_dp_info()


def get_dp_info() -> dict:
    """Get dp health info."""
    return dict(get_json(get_dp_info_endpoint()))



"""API for DP /proxy/{} endpoint."""

import json

from cmem.cmempy import config

from cmem.cmempy.api import send_request


def get(endpoint_id="default"):
    """Retrieve list of graphs from DP endpoint."""
    proxy_uri = config.get_dp_api_endpoint() + "/proxy/{}"
    return json.loads(
        send_request(proxy_uri.format(endpoint_id), method="GET").decode("utf-8")
    )



"""API for eccenca DataPlatform Label Resolution."""

import json

from cmem.cmempy import config

from cmem.cmempy.api import request


def resolve(iris, graph=None):
    """Resolve IRIs to short form labels.

    If none can be found in the data, the iri itself is used to create one.

    Possible extension: langPrefs parameter

    Args:
        iris: The list of iris to resolve.
        graph: The context/graph, in which the lookup should take place.

    Raises:
        ValueError
    """
    uri = config.get_dp_api_endpoint() + "/api/explore/titles"

    if not isinstance(iris, list):
        raise ValueError("Parameter 'iris' has to be a list of (iri) strings.")

    data = json.dumps(iris)
    headers = {"Content-type": "application/json", "Accept": "application/json"}

    params = {}
    if graph is not None:
        params["contextGraph"] = graph

    response = json.loads(
        request(uri, method="POST", headers=headers, params=params, data=data).content
    )
    return response



"""API for refreshing access conditions in DP."""

from cmem.cmempy import config

from cmem.cmempy.api import request


def get_authorization_refresh_uri():
    """Get endpoint URI for the authorization refresh."""
    return config.get_dp_api_endpoint() + "/authorization/refresh"


AUTHORIZATION_GRAPH_URI = "https://ns.eccenca.com/data/ac/"


def get():
    """GET -> Refresh access conditions."""
    return request(get_authorization_refresh_uri(), method="GET")



"""API for running SPARQL queries on DP proxy endpoints."""

from base64 import b64encode

from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get_update_endpoint():
    """Get endpoint URI pattern for a SPARQL update endpoint."""
    return config.get_dp_api_endpoint() + "/proxy/{}/update"


# pylint: disable-msg=too-many-arguments
def post(
    query,
    endpoint_id="default",
    accept="application/sparql-update",
    named_graph_uri=(),
    default_graph_uri=(),
    base64_encoded=False,
    timeout=None,
):
    """POST update / insert query."""
    headers = {"Accept": accept}
    params = {}
    uri = get_update_endpoint().format(endpoint_id)

    if base64_encoded:
        query = b64encode(query.encode())
        data = {"update": query, "base64encoded": "true"}
    else:
        data = {"update": query}

    if named_graph_uri:
        data["named-graph-uri"] = []
        for graph in named_graph_uri:
            data["named-graph-uri"].append(graph)

    if default_graph_uri:
        data["default-graph-uri"] = []
        for graph in default_graph_uri:
            data["default-graph-uri"].append(graph)

    if timeout:
        params["timeout"] = timeout

    response = send_request(
        uri=uri, method="POST", data=data, params=params, headers=headers
    )
    return response.decode("utf-8")



"""API for working with transformation tasks."""
import json

from cmem.cmempy import config

from cmem.cmempy.api import send_request


def get_task_uri():
    """Get endpoint URI pattern for a transformation task."""
    path = "/transform/tasks/{}/{}"
    return config.get_di_api_endpoint() + path


def get_transformation_task(project_name, task_name):
    """GET retrieve transformation task."""
    headers = {"Accept": "application/json"}
    response = send_request(
        get_task_uri().format(project_name, task_name), method="GET", headers=headers
    )
    return json.loads(response.decode("utf-8"))


def make_new_transformation_task(project_name, task_name, data=None):
    """PUT create transformation task."""
    headers = {"Content-Type": "application/json"}
    send_request(
        get_task_uri().format(project_name, task_name),
        method="PUT",
        data=data,
        headers=headers,
    )


def delete_transformation_task(project_name, task_name, remove_dependent_tasks=False):
    """DELETE transformation task."""
    params = {"removeDependentTasks": str(remove_dependent_tasks).lower()}
    send_request(
        get_task_uri().format(project_name, task_name), method="DELETE", params=params
    )



"""API for working with the query catalog."""

import json
import os
import re
from urllib.parse import quote
from uuid import uuid1

from pyparsing import ParseException
from rdflib.plugins.sparql import prepareQuery
from rdflib.plugins.sparql.parser import parseUpdate

from cmem.cmempy.api import get_json, request
from cmem.cmempy.config import get_cmem_base_uri, get_dp_api_endpoint
from cmem.cmempy.dp.proxy import sparql, update

# read queries are send to the normal query endpoint
QUERY_TYPES_READ = ("SELECT", "ASK", "DESCRIBE", "CONSTRUCT")
# update queries are sent to the update query endpoint
QUERY_TYPES_UPDATE_TRIPLE = ("UPDATE", "DELETE", "INSERT", "CLEAR", "LOAD")
QUERY_TYPES_UPDATE_GRAPHS = ("CREATE", "DROP", "COPY", "MOVE", "ADD")
QUERY_TYPES_UPDATE = QUERY_TYPES_UPDATE_TRIPLE + QUERY_TYPES_UPDATE_GRAPHS
# broken types
QUERY_TYPES_FAULTY = ("FAULTY", "UNKNOWN")
# all types
QUERY_TYPES = QUERY_TYPES_READ + QUERY_TYPES_UPDATE + QUERY_TYPES_FAULTY

UPDATE_CLASS = "https://vocab.eccenca.com/shui/SparqlUpdate"

QUERY_STRING = """
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX dct: <http://purl.org/dc/terms/>
PREFIX shui: <https://vocab.eccenca.com/shui/>
PREFIX dct: <http://purl.org/dc/terms/>
SELECT DISTINCT ?query ?label ?text ?type ?description ?type_class
FROM <{{GRAPH_IRI}}>
WHERE {
  ?query a ?type_class .
  FILTER (?type_class IN (shui:SparqlQuery, shui:SparqlUpdate))
  ?query rdfs:label ?label .
  ?query shui:queryText ?text .
  OPTIONAL {?query shui:queryType ?type . }
  OPTIONAL {?query dct:description ?description .}
}
"""

DEFAULT_NS = "https://ns.eccenca.com/data/queries/"
DEFAULT_GRAPH = "https://ns.eccenca.com/data/queries/"


def get_query_status():
    """Get status information of run and running queries."""
    endpoint = get_dp_api_endpoint() + "/api/admin/currentQueries"
    return get_json(endpoint)


def cancel_query(query_id: str):
    """Cancel a running query."""
    endpoint = f"{get_dp_api_endpoint()}/api/admin/currentQueries/{query_id}"
    return request(endpoint, method="DELETE")


def get_query_editor_for_uri():
    """Get query editor URI pattern for a query URI."""
    return get_cmem_base_uri() + "/query?query={}&graph={}"


def get_query_editor_for_string():
    """Get query editor URI pattern for a query string."""
    return get_cmem_base_uri() + "/query?queryString={}"


class SparqlQuery:
    """A SPARQL query with optional placeholders."""

    # pylint: disable=too-many-arguments, too-many-instance-attributes

    def __init__(
        self,
        text,
        url=None,
        label=None,
        query_type=None,
        description=None,
        origin="unknown",
        placeholder=None,
    ):
        """Initialize a SparqlQuery object."""
        self.text = text
        self.short_url = None  # will be set with self.set_url
        self.url = self.set_url(url)
        self.label = label
        self.query_type = self.set_query_type(query_type, placeholder=placeholder)
        self.description = description
        # can be one of unknown, remote or file (used for get_editor_uri)
        self.origin = origin
        self.placeholder = placeholder

    def set_url(self, url=None):
        """Set (or generate) an URL for the query."""
        if url is not None:
            self.url = url
            self.short_url = url.replace(DEFAULT_NS, ":")
        else:
            uuid = str(uuid1())
            self.url = DEFAULT_NS + uuid
            self.short_url = ":" + uuid
        return self.url

    def __str__(self):
        """Get string representation (URL)."""
        return self.url

    def set_query_type(self, query_type=None, placeholder=None):
        """Set query type."""
        self.query_type = "UNKNOWN"
        if str(query_type).upper() in QUERY_TYPES:
            self.query_type = str(query_type).upper()
        else:
            parsed_type = str(self.get_query_type(placeholder=placeholder)).upper()
            if parsed_type in QUERY_TYPES:
                self.query_type = parsed_type
        return self.query_type

    def get_query_type(self, placeholder=None):
        """Get the type of query when the placeholders are filled.

        Tries to get the type of the query - returns UNKNOWN if not possible,
        UPDATE, when an update query or a more specific READ query type
        """
        if placeholder is None:
            placeholder = {}
        algebra_names = {
            "DescribeQuery": "DESCRIBE",
            "ConstructQuery": "CONSTRUCT",
            "SelectQuery": "SELECT",
            "AskQuery": "ASK",
        }
        # prepare query text to parse
        string = self.text
        try:
            # try to fill with placeholders
            string = self.get_filled_text(placeholder)
        except ValueError:
            pass
        # 1. try parse as SPARQL query first and map the algebra type to our type
        try:
            parsed_query = prepareQuery(string)
            if str(parsed_query.algebra.name) in algebra_names:
                # return the mapped read query type
                return algebra_names[str(parsed_query.algebra.name)]
        except ParseException:
            pass
        # 2. try to parse as UPDATE query
        try:
            parseUpdate(string)
            # return the default UPDATE query type
            return "UPDATE"
        except ParseException:
            pass
        # return the UNKNOWN query type
        return "UNKNOWN"

    def get_default_accept_header(self):
        """Return the default accept header string for the query.

        return value is based on the query type and biased towards
        command line shell environment.
        """
        default_header = {
            "SELECT": "text/csv",
            "ASK": "text/csv",
            "DESCRIBE": "text/turtle",
            "CONSTRUCT": "text/turtle",
        }
        return default_header.get(self.query_type, "*")

    def get_placeholder_keys(self, text=None):
        """Get all placeholder of a query text as a set of keys."""
        if text is None:
            text = self.text
        keys = re.findall(r"{{([a-zA-Z0-9_-]+)}}", text)
        return set(keys)

    def get_filled_text(self, placeholder):
        """Replace placeholders with given values and return text.

        raises an ValueError exception if not all placeholders are filled.
        """
        text = self.text
        for key, value in placeholder.items():
            text = text.replace("{{" + key + "}}", value)
        if self.get_placeholder_keys(text):
            raise ValueError(
                "Not all placeholders filled for executing the "
                "following query:\n"
                f"- Label: {self.label!s}\n"
                f"- ID: {self.url!s}\n"
                f"- Missing parameters: {self.get_placeholder_keys(text)}"
            )
        return text

    def get_csv_results(self, placeholder=None, owl_imports_resolution=True):
        """Get results as CSV text."""
        if self.query_type != "SELECT":
            raise ValueError("Wrong query type. CSV result only supported for SELECT queries.")
        results = self.get_results(
            placeholder=placeholder,
            owl_imports_resolution=owl_imports_resolution,
            accept="text/csv",
        )
        return results

    def get_json_results(self, placeholder=None, owl_imports_resolution=True):
        """Get results as parsed json object."""
        results = self.get_results(
            placeholder=placeholder,
            owl_imports_resolution=owl_imports_resolution,
            accept="application/sparql-results+json",
        )
        return json.loads(results)

    def get_results(
        self,
        placeholder=None,
        owl_imports_resolution=True,
        base64_encoded=False,
        accept="application/sparql-results+json",
        distinct=False,
        limit=None,
        offset=None,
        timeout=None,
        replace_placeholder=True,
    ):
        """Get results as raw output from cmem.cmempy.dp.proxy.sparql."""
        if placeholder is None:
            placeholder = {}
        if self.query_type is None:
            # try to do last minute determination of type
            self.query_type = self.get_query_type(placeholder)
        if replace_placeholder:
            query_text = self.get_filled_text(placeholder)
        else:
            query_text = self.text
        if self.query_type in QUERY_TYPES_UPDATE:
            return update.post(
                query=query_text,
                accept=accept,
                base64_encoded=base64_encoded,
                timeout=timeout,
            )
        return sparql.post(
            query=query_text,
            accept=accept,
            base64_encoded=base64_encoded,
            owl_imports_resolution=owl_imports_resolution,
            distinct=distinct,
            limit=limit,
            offset=offset,
            timeout=timeout,
        )

    def get_editor_url(self, graph: str = DEFAULT_GRAPH):
        """Get query editor URI for the query."""
        if self.origin == "remote":
            return get_query_editor_for_uri().format(self.url, graph)
        if self.origin == "file":
            return get_query_editor_for_string().format(quote(self.text, safe=""))
        raise ValueError("Unknown origin, can not provide editor URL: " + self.origin)


class QueryCatalog:
    """A representation of the query catalog."""

    def __init__(self, graph: str = DEFAULT_GRAPH):
        """Initialize the catalog."""
        self.queries = None
        self.graph = graph

    def get_query(self, identifier, placeholder=None):
        """Get a query by giving an url, a short_url or a file name."""
        if os.path.isfile(identifier):
            # do not fetch the catalog in case a file is requested
            with open(identifier, encoding="UTF-8") as file_handle:
                return SparqlQuery(
                    file_handle.read(),
                    label="query from file " + identifier,
                    origin="file",
                    placeholder=placeholder,
                )
        queries = self.get_queries()
        if DEFAULT_NS + identifier[1:] in queries:
            # if the identifier comes from shell (:uuid) then we remove : and add
            # the default namespace
            return queries[DEFAULT_NS + identifier[1:]]
        if identifier in queries:
            # maybe the correct full URI is given
            return queries[identifier]
        return None

    def get_queries(self):
        """Get the query catalog as a dictionary of SparqlQuery objects."""
        if self.queries is not None:
            return self.queries
        queries = {}
        results = SparqlQuery(QUERY_STRING, query_type="SELECT").get_json_results(
            placeholder={"GRAPH_IRI": self.graph}
        )
        for query in results["results"]["bindings"]:
            description = query.get("description", {})
            description = description.get("value", None)
            query_type = query.get("type", {}).get("value", "SELECT")
            if query["type_class"]["value"] == UPDATE_CLASS:
                query_type = "UPDATE"

            queries[query["query"]["value"]] = SparqlQuery(
                query["text"]["value"],
                url=query["query"]["value"],
                label=query["label"]["value"],
                query_type=query_type,
                description=description,
                origin="remote",
            )
        self.queries = queries
        return self.queries


QUERY_CATALOG = QueryCatalog()



"""SHACL validation functions."""
from typing import Union

from cmem.cmempy import config
from cmem.cmempy.api import get_json, request

STATUS_SCHEDULED = "SCHEDULED"
STATUS_RUNNING = "RUNNING"
STATUS_FINISHED = "FINISHED"
STATUS_CANCELLED = "CANCELLED"
STATUS_ERROR = "ERROR"


def get_validation_api_endpoint() -> str:
    """Get endpoint URI for a SHACL Validation endpoint"""
    return str(config.get_dp_api_endpoint() + "/api/shacl/validation/batches")


def get_all_aggregations() -> list[dict]:
    """Get all batch validations (aggregation views)"""
    endpoint = get_validation_api_endpoint() + "/view/aggregation"
    return list(get_json(endpoint))


# pylint: disable-msg=too-many-arguments
def start(
    context_graph: str,
    shape_graph: Union[str, None] = None,
    query: Union[str, None] = None,
    result_graph: Union[str, None] = None,
    replace: bool = False,
    ignore_graph: Union[list[str], None] = None,
) -> str:
    """Start a single batch validation and return batch ID"""
    params: dict[str, Union[str, bool, list[str]]] = {"contextGraph": context_graph}
    if shape_graph:
        params["shapeGraph"] = shape_graph
    if query:
        params["selectionQuery"] = query
    if result_graph:
        params["validationResultsTargetGraph"] = result_graph
        params["replace"] = replace
    if ignore_graph:
        params["owlImportsIgnoreList"] = ignore_graph

    response = request(get_validation_api_endpoint(), params=params, method="POST")
    return str(response.headers.get("Location")).rsplit("/", maxsplit=1)[-1]


def get_aggregation(batch_id: str) -> dict:
    """Get a single batch validation (aggregation view)"""
    endpoint = get_validation_api_endpoint() + "/" + batch_id + "/view/aggregation"
    return dict(get_json(uri=endpoint))


def get(batch_id: str) -> dict:
    """Get a single batch validation"""
    endpoint = get_validation_api_endpoint() + "/" + batch_id
    return dict(get_json(uri=endpoint))


def cancel(batch_id: str) -> None:
    """Cancel a single batch validation"""
    endpoint = get_validation_api_endpoint() + "/" + batch_id
    request(uri=endpoint, method="DELETE")



"""API for working with project activities."""
import json

from cmem.cmempy import config

from cmem.cmempy.api import send_request
from cmem.cmempy.workspace.activities.projectactivity import start_project_activity


def get_activities_uri():
    """Get endpoint URI pattern for activities."""
    path = "/workspace/projects/{}/activities"
    return config.get_di_api_endpoint() + path


def get_project_activities(project_name):
    """GET retrieve project activities."""
    response = send_request(get_activities_uri().format(project_name), method="GET")
    return json.loads(response.decode("utf-8"))


# This function is kept here for backward compatibility
def start_activity(project_name, activity_name, blocking=False, data=None):
    """Kept for backward compatibility reasons, do not use."""
    return start_project_activity(
        project_name, activity_name, blocking=blocking, data=data
    )



"""cmempy configuration."""
import os

import certifi


def get_cmem_user_agent():
    """Get the CMEM_USER_AGENT config value.

    The value of user agent is send with each request.
    """
    return os.environ.get("CMEM_USER_AGENT", "unknown")


def get_cmem_base_protocol():
    """Get the CMEM_BASE_PROTOCOL config value."""
    return os.environ.get("CMEM_BASE_PROTOCOL", "http")


def get_cmem_base_domain():
    """Get the CMEM_BASE_DOMAIN config value."""
    return os.environ.get("CMEM_BASE_DOMAIN", "docker.localhost")


def get_cmem_base_uri():
    """Get the CMEM_BASE_URI config value (no ending slash)."""
    return str(
        os.environ.get(
            "CMEM_BASE_URI", f"{get_cmem_base_protocol()}://{get_cmem_base_domain()}"
        ).strip("/")
    )


def get_keycloak_base_uri():
    """Get the KEYCLOAK_BASE_URI config value (no ending slash)."""
    return str(
        os.environ.get("KEYCLOAK_BASE_URI", f"{get_cmem_base_uri()}/auth").strip("/")
    )


def get_keycloak_realm_id():
    """Get the KEYCLOAK_REALM_NAME config value (no ending slash)."""
    return str(os.environ.get("KEYCLOAK_REALM_ID", "cmem").strip("/"))


def get_ssl_verify():
    """Get the SSL_VERIFY config value."""
    ssl_verify = os.environ.get("SSL_VERIFY", "True")
    if ssl_verify.lower() == "false":
        ssl_verify = False
    else:
        ssl_verify = True
    return ssl_verify


def get_requests_ca_bundle():
    """Get the REQUESTS_CA_BUNDLE config value."""
    requests_ca_bundle = os.environ.get("REQUESTS_CA_BUNDLE", certifi.where())
    requests_ca_bundle = os.path.abspath(requests_ca_bundle)
    if not os.path.isfile(requests_ca_bundle):
        raise ValueError(
            "The value of REQUESTS_CA_BUNDLE does not point to "
            f"a CA bundle file ({requests_ca_bundle})."
        )
    return requests_ca_bundle


def get_di_api_endpoint():
    """Get the DI_API_ENDPOINT config value."""
    return os.environ.get("DI_API_ENDPOINT", f"{get_cmem_base_uri()}/dataintegration")


def get_dp_api_endpoint():
    """Get the DP_API_ENDPOINT config value."""
    return os.environ.get("DP_API_ENDPOINT", f"{get_cmem_base_uri()}/dataplatform")


def get_oauth_token_uri():
    """Get the OAUTH_TOKEN_URI config value."""
    return os.environ.get(
        "OAUTH_TOKEN_URI",
        f"{get_keycloak_base_uri()}/realms/{get_keycloak_realm_id()}/protocol/"
        "openid-connect/token",
    )


def get_oauth_authorization_uri():
    """Get the OAUTH_AUTHORIZATION_URI config value."""
    return os.environ.get(
        "OAUTH_AUTHORIZATION_URI",
        f"{get_keycloak_base_uri()}/realms/{get_keycloak_realm_id()}/protocol/"
        "openid-connect/auth",
    )


def get_oauth_grant_type():
    """Get the OAUTH_GRANT_TYPE config value."""
    valid_values = ("password", "client_credentials", "prefetched_token")
    value = os.environ.get("OAUTH_GRANT_TYPE", "client_credentials")
    if value not in valid_values:
        raise ValueError(
            f"OAUTH_GRANT_TYPE is set to {value} "
            f"which is not set valid value {valid_values}"
        )
    if value == "prefetched_token" and get_oauth_access_token() is None:
        raise ValueError(
            "OAUTH_GRANT_TYPE is set to 'prefetched_token', "
            "but no OAUTH_ACCESS_TOKEN is given."
        )
    return value


def get_oauth_access_token():
    """Get the OAUTH_ACCESS_TOKEN config value."""
    return os.environ.get("OAUTH_ACCESS_TOKEN", None)


def get_oauth_user():
    """Get the OAUTH_USER config value."""
    default_value = None
    if get_oauth_grant_type() == "password":
        default_value = "admin"
    return os.environ.get("OAUTH_USER", default_value)


def get_oauth_password():
    """Get the OAUTH_PASSWORD config value."""
    default_value = None
    if get_oauth_grant_type() == "password":
        default_value = "admin"
    return os.environ.get("OAUTH_PASSWORD", default_value)


def get_oauth_client_id():
    """Get the OAUTH_CLIENT_ID config value."""
    return os.environ.get("OAUTH_CLIENT_ID", "cmem-service-account")


def get_oauth_client_secret():
    """Get the OAUTH_CLIENT_SECRET config value."""
    return os.environ.get("OAUTH_CLIENT_SECRET", "c8c12828-000c-467b-9b6d-2d6b5e16df4a")


def get_oauth_default_credentials():
    """Get the default credentials object."""
    return {
        "grant_type": get_oauth_grant_type(),
        "username": get_oauth_user(),
        "password": get_oauth_password(),
        "client_id": get_oauth_client_id(),
        "client_secret": get_oauth_client_secret(),
    }


def is_chatty():
    """Check if API is in chatty mode. True means additional print outputs."""
    env = os.environ.get("CMEMPY_IS_CHATTY", "true")
    if env.lower() == "false":
        # silentio
        return False
    # oratio
    return True


CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
DATA_DIR = os.path.join(CURRENT_DIR, "..", "..", "tests", "fixtures_dir")



"""API for working with project variables."""
import json
from typing import Union

from requests import HTTPError

from cmem.cmempy import config

from cmem.cmempy.api import send_request
from cmem.cmempy.workspace.projects.project import get_projects


def get_variables_uri():
    """Get endpoint URI pattern for variables list."""
    path = "/api/core/variableTemplate/variables?project={}"
    return config.get_di_api_endpoint() + path


def get_variable_uri():
    """Get endpoint URI pattern for a single variable"""
    path = "/api/core/variableTemplate/variables/{}?project={}"
    return config.get_di_api_endpoint() + path


def get_variables(project_name: str) -> list[dict]:
    """GET retrieve list of variables."""
    response = send_request(get_variables_uri().format(project_name), method="GET")
    parsed = json.loads(response.decode("utf-8"))["variables"]
    returned = []
    for _ in parsed:
        returned.append(dict(_))
    return returned


def get_variable(variable_name: str, project_name: str) -> Union[dict, bool]:
    """GET retrieve a single variable.

    or False if variable was not found.
    """
    try:
        response = send_request(
            get_variable_uri().format(variable_name, project_name), method="GET"
        )
    except HTTPError as error:
        if error.response is not None and error.response.status_code == 404:
            return False
        raise error
    return dict(json.loads(response.decode("utf-8")))


def delete_variable(variable_name: str, project_name: str):
    """DELETE a single variable."""
    send_request(
        get_variable_uri().format(variable_name, project_name), method="DELETE"
    )


def create_or_update_variable(variable_name: str, project_name: str, data: dict):
    """PUT create or update a single variable."""
    headers = {"Content-Type": "application/json"}
    send_request(
        get_variable_uri().format(variable_name, project_name),
        method="PUT",
        data=json.dumps(data),
        headers=headers,
    )


def get_all_variables() -> list[dict]:
    """Get all variables of all projects."""
    variables = []
    for project_id in [_["name"] for _ in get_projects()]:
        for variable in list(get_variables(project_id)):
            variable["project"] = project_id
            variable["id"] = project_id + ":" + variable["name"]
            variables.append(variable)
    return variables



"""API for managing graphs in DP."""

import io
import os
from typing import Union

try:
    from urllib import quote  # type: ignore # Python 2.X
except ImportError:
    from urllib.parse import quote  # Python 3+

from requests.exceptions import HTTPError

from cmem.cmempy import config
from cmem.cmempy.api import get_json, request


def get_graph_uri_pattern():
    """Get endpoint URI pattern for a graph (graph store protocol)."""
    return config.get_dp_api_endpoint() + "/proxy/{}/graph?graph={}"


def _get_graph_uri(endpoint_id, graph):
    escaped_graph = quote(graph)
    return get_graph_uri_pattern().format(endpoint_id, escaped_graph)


def get(
    graph,
    endpoint_id="default",
    owl_imports_resolution=False,
    accept="application/n-triples",
    stream=False,
):
    """GET graph.

    per default, graphs are exported without imported graphs.

    Args:
        graph (str): The URI of the requested graph.
        endpoint_id (str): Optional endpoint ID (always 'default').
        owl_imports_resolution: Optional request imported graph as well.
        accept: Optional mimetype to request.
        stream: Enable streaming on the response object.

    Returns:
        requests.Response object

    """
    headers = {"Accept": accept}
    uri = (
        _get_graph_uri(endpoint_id, graph)
        + "&owlImportsResolution="
        + str(owl_imports_resolution).lower()
    )
    return request(uri, method="GET", headers=headers, stream=stream)


def get_streamed(
    graph,
    endpoint_id="default",
    owl_imports_resolution=False,
    accept="application/n-triples",
):
    """GET graph (streamed).

    same as get

    Args:
        graph (str): The URI of the requested graph.
        endpoint_id (str): Optional endpoint ID (always 'default').
        owl_imports_resolution: Optional request imported graph as well.
        accept: Optional mimetype to request.

    Returns:
        requests.Response object

    """
    return get(
        graph,
        endpoint_id=endpoint_id,
        owl_imports_resolution=owl_imports_resolution,
        accept=accept,
        stream=True,
    )


def delete(graph, endpoint_id="default", silent=True):
    """DELETE graph."""
    uri = _get_graph_uri(endpoint_id, graph)
    response = request(uri, method="DELETE", raise_for_status=False)
    if response.status_code == 404 and silent:
        return response
    response.raise_for_status()
    return response


def post(graph, file, endpoint_id="default", replace=False):
    """Upload graph.

    Add the content of triple to a remote graph or replace the remote graph
    with the content of a triple file.

    Args:
        graph (str): The URI of the remote graph.
        file: The triple file
        endpoint_id (str): Optional endpoint ID (always 'default').
        replace (bool): add (False) or replace (True)

    Returns:
        requests.Response object

    """
    uri = _get_graph_uri(endpoint_id, graph) + "&replace=" + str(replace).lower()
    with open(file, "rb") as file_handle:
        return request(uri, method="POST", files={"file": (os.path.basename(file), file_handle)})


def post_streamed(
    graph,
    file: Union[str, io.BytesIO],
    endpoint_id="default",
    replace=False,
    content_type="text/turtle",
    content_encoding=None,
):  # pylint: disable=too-many-arguments
    """Upload graph (streamed).

    Add the content of triple to a remote graph or replace the remote graph
    with the content of a triple file.

    Args:
        graph (str): The URI of the remote graph.
        file: The triple file
        endpoint_id (str): Optional endpoint ID (always 'default').
        replace (bool): add (False) or replace (True)
        content_type (str): mime type of the file to post (default is turtle)

    Returns:
        requests.Response object

    """
    uri = _get_graph_uri(endpoint_id, graph) + "&replace=" + str(replace).lower()
    headers = {"Content-Type": content_type}
    if content_encoding:
        headers["Content-Encoding"] = content_encoding
    # https://2.python-requests.org/en/master/user/advanced/#streaming-uploads
    if isinstance(file, io.BytesIO):
        response = request(uri, method="POST", headers=headers, data=file, stream=True)
    else:
        with open(file, "rb") as opened_file:
            response = request(uri, method="POST", headers=headers, data=opened_file, stream=True)
    return response


def put(graph, rdf_data, rdf_serialization, endpoint_id="default"):
    """PUT graph."""
    headers = {"Content-Type": rdf_serialization}
    uri = _get_graph_uri(endpoint_id, graph)
    return request(uri, method="PUT", data=rdf_data, headers=headers)


def get_graphs_list():
    """Get a list of graph descriptions the user is allowed to access.

    Returns:
        parsed json response
    Examples
        {
        "assignedClasses": [
            "http://rdfs.org/ns/void#Dataset"
        ],
        "diProjectGraph": false,
        "graphProperies": [
            "https://vocab.eccenca.com/shui/isSystemResource"
        ],
        "iri": "https://ns.eccenca.com/data/queries/",
        "label": {
            "fromIri": false,
            "graph": "https://ns.eccenca.com/data/queries/",
            "iri": "https://ns.eccenca.com/data/queries/",
            "lang": "en",
            "title": "CMEM Query Catalog",
            "when": "16:35:40.353"
        },
        "projectInternal": false,
        "systemResource": true,
        "writeable": true
    }

    """
    endpoint = config.get_dp_api_endpoint() + "/graphs/list"
    return get_json(endpoint)


def get_graph_imports(graph):
    """Get the list of graphs, which a graph imports, resolved transitively.

    Args:
        graph (str): The IRI of the graph

    Returns:
        list of graph IRIs which this graph imports

    """
    endpoint = config.get_dp_api_endpoint() + "/graphs/imports"
    params = {"graph": graph}
    return get_json(endpoint, params=params)


def get_graph_import_tree(graph):
    """Get the graph imported graphs, which this graph includes.

    Args:
        graph (str): The IRI of the graph

    Returns:
        structure of graph IRIs which this graph imports

    """
    endpoint = config.get_dp_api_endpoint() + "/graphs/importTree"
    params = {"graph": graph}
    try:
        response = get_json(endpoint, params=params)
    except HTTPError:
        return {graph: []}
    if response == {}:
        return {graph: []}
    return get_json(endpoint, params=params)



"""Workspace API."""
from cmem.cmempy import config

from cmem.cmempy.api import get_json, send_request


def get_reload_uri():
    """Get URL for workspace reload."""
    return config.get_di_api_endpoint() + "/workspace/reload"


def get_reload_prefixes_uri():
    """Get URL for prefix reload."""
    return config.get_di_api_endpoint() + "/workspace/reloadPrefixes"


def get_update_global_vocabulary_cache_uri():
    """Get URL for prefix reload."""
    di_base = config.get_di_api_endpoint()
    return di_base + "/workspace/updateGlobalVocabularyCache"


def reload_workspace():
    """Reload the workspace from the backend."""
    return send_request(get_reload_uri(), method="POST")


def get_task_plugins():
    """
    Get a list of plugins that can be created as workspace tasks.

    A list of plugins that can be created as workspace tasks, e.g. datasets,
    transform tasks etc. The result of this endpoint only contains meta data
    of the plugin, i.e. title, description and categories.
    To fetch the schema details of a specific plugin use the /plugin endpoint.
    """
    return get_json(
        config.get_di_api_endpoint() + "/api/core/taskPlugins", method="GET"
    )


def get_task_plugin_description(plugin_id):
    """Get The plugin description of a specific plugin."""
    path = f"/api/core/plugins/{plugin_id}"
    return get_json(config.get_di_api_endpoint() + path, method="GET")


def reload_prefixes():
    """Reload the workspace from the backend."""
    return send_request(get_reload_prefixes_uri(), method="POST")


def update_global_vocabulary_cache(iri):
    """Update the global vocabulary cache."""
    data = '{"iri":"' + iri + '"}'
    headers = {"Content-Type": "text/json"}
    return send_request(
        get_update_global_vocabulary_cache_uri(),
        method="POST",
        data=data,
        headers=headers,
    )



"""API for managing conditions in DP."""
import json
from typing import Optional, Union

from cmem.cmempy import config

from cmem.cmempy.api import request


def get_authorization_uri():
    """Get graph URI of the authorization conditions graph."""
    path = "/api/authorization"
    return config.get_dp_api_endpoint() + path


def get_acls(page: int = 0):
    """GET access conditions."""
    return request(get_authorization_uri(), method="GET", params={"page": page})


def fetch_all_acls() -> list:
    """Fetch all the access conditions"""
    acls: list[dict] = []
    page: int = 0
    while True:
        result = get_acls(page=page).json()
        acls += result["content"]
        # CMEM-6362: until 24.3, we had a serialization, where 'last' was explicit
        # last = True -> last page (old page model)
        # last = False -> not last page (old page model)
        # KeyError -> new page model, lets check number and totalPages
        last_indicator = False
        try:
            last_indicator = result["last"]
        except KeyError:
            number = int(result["page"]["number"]) + 1  # starts with 0
            total = int(result["page"]["totalPages"])
            if number >= total:
                last_indicator = True
        if last_indicator is True:
            break
        page += 1
    return acls


# pylint: disable-msg=too-many-arguments
def update_access_condition(
    iri,
    name=None,
    description=None,
    user=None,
    groups=None,
    read_graphs=None,
    write_graphs=None,
    actions=None,
    read_graph_patterns=None,
    write_graph_patterns=None,
    action_patterns=None,
    query=None,
):
    """PUT access conditions."""
    payload = get_access_condition_by_iri(iri=iri).json()
    if name:
        payload["name"] = name
    if description:
        payload["comment"] = description
    if user:
        payload["requiresAccount"] = user
    if groups:
        payload["requiresGroup"] = groups
    if read_graphs:
        payload["readableGraphs"] = read_graphs
    if write_graphs:
        payload["writableGraphs"] = write_graphs
    if actions:
        payload["allowedActions"] = actions
    if read_graph_patterns:
        payload["grantReadPatterns"] = read_graph_patterns
    if write_graph_patterns:
        payload["grantWritePatterns"] = write_graph_patterns
    if action_patterns:
        payload["grantAllowedActions"] = action_patterns
    if query:
        payload["dynamicAccessConditionQuery"] = query

    payload = json.dumps(payload)
    content_type = "application/json"
    headers = {"accept": content_type, "Content-Type": content_type}
    return request(get_authorization_uri(), method="PUT", data=payload, headers=headers)


# pylint: disable-msg=too-many-arguments, too-many-locals
def create_access_condition(
    name: str,
    static_id: Optional[str] = None,
    description: Optional[str] = None,
    user: Optional[str] = None,
    groups: Optional[list[str]] = None,
    read_graphs: Optional[list[str]] = None,
    write_graphs: Optional[list[str]] = None,
    actions: Optional[list[str]] = None,
    read_graph_patterns: Optional[list[str]] = None,
    write_graph_patterns: Optional[list[str]] = None,
    action_patterns: Optional[list[str]] = None,
    query: Optional[str] = None,
):
    """POST access conditions."""
    payload = {
        "staticId": static_id,
        "name": name,
        "comment": description,
        "requiresAccount": user,
        "requiresGroup": groups,
        "readableGraphs": read_graphs,
        "writableGraphs": write_graphs,
        "allowedActions": actions,
        "grantReadPatterns": read_graph_patterns,
        "grantWritePatterns": write_graph_patterns,
        "grantAllowedActions": action_patterns,
        "dynamicAccessConditionQuery": query,
    }
    content_type = "application/json"
    data = json.dumps(payload).encode("utf-8")
    headers = {"accept": content_type, "Content-Type": content_type}
    return request(get_authorization_uri(), method="POST", headers=headers, data=data)


def delete_access_condition(iri):
    """DELETE access condition."""
    params = {"resource": iri}
    return request(get_authorization_uri(), method="DELETE", params=params)


def get_users():
    """GET users."""
    uri = get_authorization_uri() + "/users"
    return request(uri, method="GET")


def review_graph_rights(
        account_iri: Union[str, None] = None,
        group_iris: Union[list[str], None] = None
):
    """Review access rights for a given account and group."""
    uri = get_authorization_uri() + "/review"

    # query parameters
    params: dict[str, Union[str, list[str], None]] = {
        "accountIri": account_iri,
        "groupIris": group_iris
    }
    return request(uri, method="GET", params=params)


def get_access_condition_by_iri(iri: str):
    """Get access condition by iri."""
    uri = get_authorization_uri() + "/iri"
    params = {"resource": iri}
    return request(uri, method="GET", params=params)


def get_groups():
    """GET groups."""
    uri = get_authorization_uri() + "/groups"
    return request(uri, method="GET")


def get_actions():
    """GET actions."""
    uri = get_authorization_uri() + "/actions"
    return request(uri, method="GET")



"""Send HTTP requests to CMEM API."""

import json

import requests

from . import config


def get_access_token(_oauth_token_uri=None, _oauth_credentials=None):
    """Return access token string.

    This is a wrapper for get_token to get only the string.
    It does NOT fetch a new token in case a 'prefetched_token" grant type.
    """
    if config.get_oauth_grant_type() == "prefetched_token":
        return config.get_oauth_access_token()
    return get_token(_oauth_token_uri, _oauth_credentials)["access_token"]


def get_token(_oauth_token_uri=None, _oauth_credentials=None):
    """Return auth token JSON structure.

    Auth token format is:
    {
        'token_type': 'bearer',
        'refresh_token': '923a28d6-470d-4afd-9b91-0c952a29a423',
        'expires_in': 43199,
        'access_token': '55c21f4e-9529-4399-aaf1-128acb9da4a1',
        'scope': 'read'
    }
    """
    if not _oauth_credentials:
        _oauth_credentials = config.get_oauth_default_credentials()
    if not _oauth_token_uri:
        _oauth_token_uri = config.get_oauth_token_uri()
    post_req = _request("post", _oauth_token_uri, data=_oauth_credentials)
    post_req.raise_for_status()
    response = post_req.content.decode("utf-8")
    try:
        parsed_json = json.loads(response)
    except json.decoder.JSONDecodeError:
        # pylint: disable=raise-missing-from
        raise ValueError(f"Token response is not proper JSON. Response was:\n{response}")
    return parsed_json


# pylint: disable-msg=too-many-arguments
def send_request(
    uri,
    method="GET",
    data=None,
    _json=None,
    headers=None,
    params=None,
    files=None,
    raise_for_status=True,
):
    """Return HTML content for request directly."""
    return request(
        uri,
        method=method,
        data=data,
        _json=_json,
        headers=headers,
        params=params,
        files=files,
        raise_for_status=raise_for_status,
    ).content


# pylint: disable-msg=too-many-arguments
def get_json(
    uri,
    method="GET",
    data=None,
    _json=None,
    headers=None,
    params=None,
    files=None,
    raise_for_status=True,
):
    """Return parsed JSON content for request directly."""
    response = request(
        uri,
        method=method,
        data=data,
        _json=_json,
        headers=headers,
        params=params,
        files=files,
        raise_for_status=raise_for_status,
    ).content
    return json.loads(response.decode("utf-8"))


def _request(method=None, url=None, **kwargs):
    # pylint: disable=missing-timeout
    return requests.api.request(method, url, verify=config.get_ssl_verify(), **kwargs)


# pylint: disable-msg=too-many-arguments
def request(
    uri,
    method="GET",
    data=None,
    _json=None,
    headers=None,
    params=None,
    files=None,
    stream=False,
    raise_for_status=True,
):
    """Return the raw response object."""
    if headers is None:
        headers = {}
    headers["Authorization"] = f"Bearer {get_access_token()!s}"
    headers["User-Agent"] = config.get_cmem_user_agent()
    if method == "GET":
        response = _request("get", uri, headers=headers, stream=stream, params=params)
    elif method == "POST":
        response = _request(
            "post",
            uri,
            headers=headers,
            data=data,
            params=params,
            files=files,
            stream=stream,
        )
    elif method == "PUT":
        if config.is_chatty():
            print(headers)
        response = _request(
            "put",
            uri,
            headers=headers,
            data=data,
            params=params,
            files=files,
            stream=stream,
        )
    elif method == "DELETE":
        response = _request("delete", uri, headers=headers, params=params, stream=stream)
    elif method == "PATCH":
        response = _request(
            "patch",
            uri,
            headers=headers,
            data=data,
            params=params,
            files=files,
            stream=stream,
        )
    else:
        raise requests.exceptions.HTTPError("Unknown HTTP request method")
    if response.status_code not in [
        requests.status_codes.codes.ok,
        requests.status_codes.codes.no_content,
        requests.status_codes.codes.created,
    ]:
        if config.is_chatty():
            print(f"Request failed for URL: {response.url} " f"{response.request.method}")
            print(response.content.decode("utf-8"))
            print(response.status_code)
    if raise_for_status:
        response.raise_for_status()
    return response



"""API methods for working with single datasets."""
import json

from requests_toolbelt.multipart.encoder import MultipartEncoder

from cmem.cmempy import config
from cmem.cmempy.api import send_request, request


def get_dataset_uri():
    """Get endpoint URI pattern for a dataset."""
    path = "/workspace/projects/{}/datasets/{}"
    return config.get_di_api_endpoint() + path


def get_dataset_file_uri():
    """Get the endpoint URI for a dataset file."""
    return get_dataset_uri() + "/file"


def get_dataset(project_name, dataset_name):
    """GET retrieve single dataset."""
    headers = {"Accept": "application/json"}
    response = send_request(
        get_dataset_uri().format(project_name, dataset_name),
        method="GET",
        headers=headers,
    )
    return json.loads(response.decode("utf-8"))


def make_new_dataset(
    project_name, dataset_name, dataset_type, parameters, autoconfigure
):
    """PUT create new dataset (deprecated)."""
    data = {
        "id": f"{dataset_name}",
        "type": f"{dataset_type}",
        "parameters": parameters,
    }
    data = json.dumps(data).encode("utf-8")
    headers = {"Content-Type": "application/json"}
    params = None
    if autoconfigure:
        params = {"autoConfigure": "true"}
    send_request(
        get_dataset_uri().format(project_name, dataset_name),
        method="PUT",
        data=data,
        headers=headers,
        params=params,
    )


def delete_dataset(project_name, dataset_name):
    """DELETE remove existing dataset."""
    send_request(get_dataset_uri().format(project_name, dataset_name), method="DELETE")


# pylint: disable=too-many-arguments
def create_dataset(
    project_id,
    dataset_type,
    dataset_id=None,
    parameter=None,
    metadata=None,
    read_only=False,
    uri_property=None,
):
    """Create a dataset.

    In difference to make_new_dataset, this uses the task API and does
    not enforce an ID in advance. Also, it allows for metadata parameters.
    """
    if parameter is None:
        parameter = {}
    if metadata is None:
        metadata = {}
    # add needed base task data
    data = {
        "taskType": "Dataset",
        "type": dataset_type,
        "parameters": parameter,
        "metadata": metadata,
        "readOnly": read_only,
        "uriProperty": uri_property,
    }
    # add the optional task / dataset ID
    if dataset_id:
        data["id"] = dataset_id
    params = None
    path = f"/workspace/projects/{project_id}/tasks"
    return send_request(
        config.get_di_api_endpoint() + path,
        method="POST",
        data=json.dumps(data).encode("utf-8"),
        headers={"Content-Type": "application/json"},
        params=params,
    )


def post_resource(project_id, dataset_id, file_resource=None, resource_url=None):
    """
    Post a resource to a dataset.

    If the dataset resource already exists, posting a new resource will replace it.

    Args:
        project_id (str): The ID of the project.
        dataset_id (str): The ID of the dataset.
        file_resource (io Binary Object, optional): The file resource to be uploaded.
        resource_url (str, optional): The URL of the resource to be uploaded.

    Returns:
        Response: The response from the request.

    """

    if not file_resource and not resource_url:
        raise ValueError("Parameter file_resource or resource_url is needed.")

    endpoint = get_dataset_file_uri().format(project_id, dataset_id)
    if file_resource:
        with file_resource as file:
            response = request(
                endpoint,
                method="PUT",
                stream=True,
                data=file,
            )
        return response

    multipart_encoder = MultipartEncoder(fields=[("resource-url", resource_url)])
    headers = {"Content-Type": multipart_encoder.content_type}
    return request(
        endpoint,
        method="PUT",
        headers=headers,
        data=multipart_encoder,
        stream=True,
    )


def get_resource(project_id, dataset_id):
    """Get a resource"""
    endpoint = get_dataset_file_uri().format(project_id, dataset_id)
    response = send_request(
        endpoint,
        method="GET",
    )
    return response.decode("utf-8")


def update_dataset(
    project_id,
    dataset_id,
    parameters=None,
    metadata=None,
    read_only=False,
    uri_property=None,
):
    """Update a dataset"""
    if parameters is None:
        parameters = {}
    if metadata is None:
        metadata = {}
    data = {
        "metadata": metadata,
        "data": {
            "readOnly": read_only,
            "uriProperty": uri_property,
            "parameters": parameters,
        },
    }
    path = f"/workspace/projects/{project_id}/tasks/{dataset_id}"
    return send_request(
        config.get_di_api_endpoint() + path,
        method="PATCH",
        data=json.dumps(data).encode("utf-8"),
        headers={"Content-Type": "application/json"},
    )
